#pragma -xO
      character*1 line(80)
      real*8 pi, pi8
      parameter (max = 30)
      call pi8 (pi)
      do 1 i = 1, max
      x = 2 * (i - 0.5) / max * pi
      y = 39 * (sin (x) + 1)
      do 2 j = 1, 80
    2 line(j) = ' '
      line(1 + int (y + 0.5)) = '*' 
    1 write (*, *) line
      end
