#pragma -xO
      PARAMETER (NS=3)
      CHARACTER*16 LINE(NS) 
      DATA LINE(1) / '   1   1' / 
      DATA LINE(2) / '   2   2' / 
      DATA LINE(3) / '   3   3' / 
      READ (LINE, FMT='(2I4)', REC=2) M, N 
      PRINT *, M, N 
      READ (LINE, FMT='(2I4)') M, N 
      PRINT *, M, N 
      DECODE (15, 10, LINE) M, N 
   10 FORMAT (2I4)
      PRINT *, M, N 
      END 
