#pragma -xO
c 'E' or 'e',   DLAMCH := eps
c 'S' or 's ,   DLAMCH := sfmin
c 'B' or 'b',   DLAMCH := base
c 'P' or 'p',   DLAMCH := eps*base
c 'N' or 'n',   DLAMCH := t
c 'R' or 'r',   DLAMCH := rnd
c 'M' or 'm',   DLAMCH := emin
c 'U' or 'u',   DLAMCH := rmin
c 'L' or 'l',   DLAMCH := emax
c 'O' or 'o',   DLAMCH := rmax
      real*8 dlamch
      write (*, *) dlamch('e')
      write (*, *) dlamch('s')
      write (*, *) dlamch('b')
      write (*, *) dlamch('p')
      write (*, *) dlamch('n')
      write (*, *) dlamch('r')
      write (*, *) dlamch('m')
      write (*, *) dlamch('u')
      write (*, *) dlamch('l')
      write (*, *) dlamch('o')
      end
