#pragma -xO
      intrinsic qsin, xsin, qasin, xasin, qexp, xexp
      print *, 'Test sin'
      call test(qsin, xsin)
      print *, 'Test asin'
      call test(qasin, xasin)
      print *, 'Test exp'
      call test(qexp, xexp)
      end

      subroutine test (fun16, fun32)
      parameter (n = 25000)
      real drand48*8, x, x16*16, x32*32, f1*16, f2*16, fun16*16, fun32*32
      do i = 1, n
         x = drand48()
         x16 = x
         x32 = x
         f1 = fun16(x16)
         f2 = fun32(x32)
         if (abs(f1 - f2) > 1e-9) then
            write (*, *) f1, f2
         endif
      end do
      end
