#pragma -xO
      program a
      implicit real*8 (a-h, o-z)
      write (*, *) wapr (0.0d0, 0, nerr, 0)
      write (*, *) wapr (-dexp (-1.0d0), 0, nerr, 0)
      write (*, *) wapr (dexp (1.0d0), 0, nerr, 0)
      write (*, *) wapr (dexp (1.0d0 + dexp (1.0d0)), 0, nerr, 0)
      write (*, *) wapr (32 * dlog(2), 0, nerr, 0)
      end
