#pragma -xO
      program random
      parameter (n = 10000000)
      real*8 rnd, sum
      sum = 0
      do i = 1, n
      call rand48 (j, k, rnd)
      sum = sum + rnd
      end do
      write (*, *) sum / n
      end
